﻿using System.Linq.Expressions;
using Drive.Common.Domain;
using Microsoft.EntityFrameworkCore;

namespace Drive.Common.Data
{
    public interface IAggregateRootLoader
    {
        Task<TAggregate> Load<TAggregate, TRecord>(Guid id)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : class, IAggregateRootRecord;

        Task<IEnumerable<TAggregate>> Load<TAggregate, TRecord>(IEnumerable<Guid> ids)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : class, IAggregateRootRecord;

        Task<IEnumerable<TAggregate>> Load<TAggregate, TRecord>(Expression<Func<TRecord, bool>> predicate)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : class, IAggregateRootRecord;

        Task<TAggregate> Load<TAggregate, TRecord>(int id)
            where TAggregate : ILegacyAggregateRoot<TRecord>
            where TRecord : class, ILegacyAggregateRootRecord;
    }

    public class AggregateRootLoader(DriveDbContext context) : IAggregateRootLoader
    {
        private readonly DriveDbContext _context = context;

        public async Task<TAggregate> Load<TAggregate, TRecord>(Guid id)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : class, IAggregateRootRecord
        {
            TRecord record = await _context
                .Set<TRecord>()
                .FindAsync(id)
                .ConfigureAwait(false)
                ?? throw new ArgumentOutOfRangeException(nameof(id));

            TAggregate result = (TAggregate?)Activator.CreateInstance(typeof(TAggregate), true)!;
            result.Restore(record);
            return result;
        }

        public async Task<IEnumerable<TAggregate>> Load<TAggregate, TRecord>(IEnumerable<Guid> ids)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : class, IAggregateRootRecord
        {
            IEnumerable<TRecord> records = await _context
                .Set<TRecord>()
                .Where(x => ids.Contains(x.Id))
                .ToListAsync()
                .ConfigureAwait(false);

            List<TAggregate> aggregates = [];
            foreach (TRecord record in records)
            {
                TAggregate aggregate = (TAggregate)Activator.CreateInstance(typeof(TAggregate), true)!;
                aggregate.Restore(record);
                aggregates.Add(aggregate);
            }

            return aggregates;
        }

        public async Task<IEnumerable<TAggregate>> Load<TAggregate, TRecord>(Expression<Func<TRecord, bool>> predicate)
            where TAggregate : IAggregateRoot<TRecord>
            where TRecord : class, IAggregateRootRecord
        {
            IEnumerable<TRecord> records = await _context
                .Set<TRecord>()
                .Where(predicate)
                .ToListAsync()
                .ConfigureAwait(false);

            List<TAggregate> aggregates = [];
            foreach (TRecord record in records)
            {
                TAggregate aggregate = (TAggregate)Activator.CreateInstance(typeof(TAggregate), true)!;
                aggregate.Restore(record);
                aggregates.Add(aggregate);
            }

            return aggregates;
        }

        public async Task<TAggregate> Load<TAggregate, TRecord>(int id)
            where TAggregate : ILegacyAggregateRoot<TRecord>
            where TRecord : class, ILegacyAggregateRootRecord
        {
            TRecord record = await _context
                .Set<TRecord>()
                .FindAsync(id)
                .ConfigureAwait(false)
                ?? throw new ArgumentOutOfRangeException(nameof(id));

            TAggregate result = (TAggregate)Activator.CreateInstance(typeof(TAggregate), true)!;
            result.Restore(record);
            return result;
        }
    }
}
