﻿using Drive.Common.Domain;

namespace Drive.Common.Data
{
    public interface IAggregateRootPersister
    {
        Task Persist<TRecord>(
            IAggregateRoot<TRecord> aggregate,
            DriveContext? context = null) where TRecord : class, IAggregateRootRecord;
        Task Persist<TRecord>(
            ILegacyAggregateRoot<TRecord> aggregate,
            DriveContext? context = null) where TRecord : class, ILegacyAggregateRootRecord;
        Task Persist<TRecord>(IEnumerable<IAggregateRoot<TRecord>> aggregates) where TRecord : class, IAggregateRootRecord;
        Task Delete<TRecord>(IEnumerable<IAggregateRoot<TRecord>> aggregates) where TRecord : class, IAggregateRootRecord;
        Task Delete<TRecord>(IAggregateRoot<TRecord> aggregate) where TRecord : class, IAggregateRootRecord;
    }

    public class AggregateRootPersister(
        DriveDbContext context,
        IClockProvider clockProvider)
        : IAggregateRootPersister
    {
        private readonly DriveDbContext _context = context;
        private readonly IClockProvider _clockProvider = clockProvider;

        public async Task Persist<TRecord>(
            IAggregateRoot<TRecord> aggregate,
            DriveContext? context = null) where TRecord : class, IAggregateRootRecord
        {
            TRecord record = aggregate.GetRecord();
            IEnumerable<object> events = aggregate.GetEvents();

            if (context != null && record is IAuditedAggregateRoot root)
            {
                root.ModifiedById = context.DriveUserId;
            }

            if (!_context.Set<TRecord>().Local.Any(x => x == record))
            {
                record.ModifiedAt = record.CreatedAt;
                _context.Set<TRecord>().Add(record);
            }
            else if (_context.ChangeTracker.HasChanges())
                record.ModifiedAt = _clockProvider.UtcNow;

            await _context
                .SaveChangesAsync()
                .ConfigureAwait(false);
        }

        public async Task Persist<TRecord>(
           IEnumerable<IAggregateRoot<TRecord>> aggregates) where TRecord : class, IAggregateRootRecord
        {
            foreach (IAggregateRoot<TRecord> aggregate in aggregates)
            {
                TRecord record = aggregate.GetRecord();
                IEnumerable<object> events = aggregate.GetEvents();

                if (!_context.Set<TRecord>().Local.Any(x => x == record))
                {
                    record.ModifiedAt = record.CreatedAt;
                    _context.Set<TRecord>().Add(record);
                }
                else if (_context.ChangeTracker.HasChanges())
                    record.ModifiedAt = _clockProvider.UtcNow;
            }

            await _context
                .SaveChangesAsync()
                .ConfigureAwait(false);
        }

        public async Task Persist<TRecord>(
            ILegacyAggregateRoot<TRecord> aggregate,
            DriveContext? context = null) where TRecord : class, ILegacyAggregateRootRecord
        {
            TRecord record = aggregate.GetRecord();
            var events = aggregate.GetEvents();

            if (context != null && record is ILegacyAuditedAggregateRoot root)
            {
                root.ModifiedById = context.DriveUserId;
            }

            if (!_context.Set<TRecord>().Local.Any(x => x == record))
            {
                record.ModifiedAt = record.CreatedAt;
                _context.Set<TRecord>().Add(record);
            }
            else if (_context.ChangeTracker.HasChanges())
                record.ModifiedAt = _clockProvider.UtcNow;

            await _context
                .SaveChangesAsync()
                .ConfigureAwait(false);
        }
        public async Task Delete<TRecord>(
            IAggregateRoot<TRecord> aggregate) where TRecord : class, IAggregateRootRecord
        {
            TRecord record = aggregate.GetRecord();

            _context.Set<TRecord>().Remove(record);

            await _context
                .SaveChangesAsync()
                .ConfigureAwait(false);
        }

        public async Task Delete<TRecord>(
            IEnumerable<IAggregateRoot<TRecord>> aggregates) where TRecord : class, IAggregateRootRecord
        {
            foreach (IAggregateRoot<TRecord> aggregate in aggregates)
            {
                TRecord record = aggregate.GetRecord();

                _context.Set<TRecord>().Remove(record);
            }

            await _context
                .SaveChangesAsync()
                .ConfigureAwait(false);
        }
    }
}
