﻿using System.Data;
using System.Data.Common;
using System.Reflection;
using Drive.Common.Data;
using Microsoft.Data.SqlClient;
using Microsoft.EntityFrameworkCore;
using Microsoft.EntityFrameworkCore.Metadata;
using Microsoft.EntityFrameworkCore.Metadata.Conventions;
using Xunit;

namespace Drive.Common.Testing.Database
{
    public abstract class DatabaseTest : IDisposable
    {
        private readonly string _serverConnectionString = @"Data Source=(LocalDB)\MSSQLLocalDB";

        private readonly string _databaseName;
        private readonly string _connectionString;

        private readonly DbContextOptionsBuilder _dbContextOptions;

        public DatabaseTest(string? serverConnectionString = null)
        {
            _serverConnectionString = serverConnectionString ?? _serverConnectionString;

            Type collectionDefinition = typeof(ICollectionFixture<>).MakeGenericType(GetType());

            Type? collection = AppDomain.CurrentDomain
                                        .GetAssemblies()
                                        .Where(x => !string.IsNullOrWhiteSpace(x.FullName) && x.FullName.Contains("Drive"))
                                        .SelectMany(x => x.ExportedTypes)
                                        .Where(x => collectionDefinition.IsAssignableFrom(x))
                                        .FirstOrDefault();

            string name = collection?.Assembly.GetName().Name ?? Assembly.GetExecutingAssembly().GetName().Name!;

            _databaseName = $"{name}_{GetType().Name}".Replace(".", "-");

            DropDatabase();
            CreateDatabase();
            _connectionString = $"{_serverConnectionString};Initial Catalog={_databaseName};Pooling=false";

            ModelBuilder modelBuilder = new(SqlServerConventionSetBuilder.Build());
            AddMappings(modelBuilder);
            IModel model = modelBuilder.FinalizeModel();

            InitializeFromContext(model);
            CreateAuxiliaryObjects();

            _dbContextOptions = new();
            _dbContextOptions.UseModel(model);

            AddInitialTestData();
        }

        protected virtual void CreateAuxiliaryObjects() { }
        protected virtual void AddMappings(ModelBuilder modelBuilder) { }
        protected virtual void AddInitialTestData() { }

        public SqlConnection OpenConnection(string? connectionString = null)
        {
            SqlConnection connection = new(connectionString ?? _connectionString);
            connection.Open();

            return connection;
        }

        public TestSqlStorage OpenSqlStorage()
        {
            SqlConnection connection = new(_connectionString);
            connection.Open();

            return new TestSqlStorage(connection);
        }

        public DriveDbContext OpenDbContext(
            DbConnection connection,
            DbTransaction? transaction = null)
        {
            _dbContextOptions.UseSqlServer(connection, contextOwnsConnection: false, x => x.UseNodaTime().UseQuerySplittingBehavior(QuerySplittingBehavior.SplitQuery));

            DriveDbContext dbContext = new(_dbContextOptions.Options);
            if (transaction is not null)
                dbContext.Database.UseTransaction(transaction);

            return dbContext;
        }

        public void ResetData()
        {
            using (SqlConnection connection = OpenConnection())
            using (SqlTransaction transaction = connection.BeginTransaction())
            using (SqlCommand command = connection.CreateCommand())
            {
                command.Transaction = transaction;
                command.CommandType = CommandType.StoredProcedure;
                command.CommandText = "sp_MSForEachTable";

                // disable referential integrity
                command.Parameters.Clear();
                command.Parameters.Add(new SqlParameter("@command1", "ALTER TABLE ? NOCHECK CONSTRAINT ALL"));
                command.ExecuteNonQuery();

                command.Parameters.Clear();
                command.Parameters.Add(new SqlParameter("@command1", "SET QUOTED_IDENTIFIER ON; DELETE FROM ?"));
                command.ExecuteNonQuery();

                // enable referential integrity again
                command.Parameters.Clear();
                command.Parameters.Add(new SqlParameter("@command1", "ALTER TABLE ? WITH CHECK CHECK CONSTRAINT ALL"));
                command.ExecuteNonQuery();

                transaction.Commit();
            }

            AddInitialTestData();
        }

        private void CreateDatabase()
        {
            using SqlConnection connection = OpenConnection(_serverConnectionString);
            using SqlCommand command = connection.CreateCommand();

            command.CommandText = $@"
CREATE DATABASE [{_databaseName}]
ON PRIMARY      (NAME=[{_databaseName}_data], FILENAME = 'C:\temp\{_databaseName}_data.mdf')
LOG ON          (NAME=[{_databaseName}_log], FILENAME = 'C:\temp\{_databaseName}_log.ldf')";
            command.ExecuteNonQuery();
        }

        private void DropDatabase()
        {
            using SqlConnection connection = OpenConnection(_serverConnectionString);
            using SqlCommand command = connection.CreateCommand();

            command.CommandText = $@"
IF EXISTS (SELECT name FROM sys.databases WHERE name = N'{_databaseName}')
    BEGIN
        ALTER DATABASE [{_databaseName}] SET SINGLE_USER WITH ROLLBACK IMMEDIATE;
        DROP DATABASE [{_databaseName}];
    END";
            command.ExecuteNonQuery();
        }

        private void InitializeFromContext(IModel model)
        {
            DbContextOptionsBuilder dbContextOptions = new();
            dbContextOptions.UseModel(model);
            dbContextOptions.UseSqlServer(_connectionString, x => x.UseNodaTime());

            using DriveDbContext dbContext = new(dbContextOptions.Options);
            dbContext.Database.EnsureCreated();
        }

        public void Dispose()
        {
            DropDatabase();
        }
    }

    public class TestSqlStorage : ISqlStorage, IDisposable
    {
        internal TestSqlStorage(SqlConnection connection)
        {
            Connection = connection;
            Transaction = Connection.BeginTransaction();
        }

        public DbConnection Connection { get; private set; }

        public DbTransaction Transaction { get; private set; }

        public void Dispose()
        {
            Transaction.Dispose();
            Connection.Dispose();
        }
    }
}